/* This file is not part of the entry as such, but provides an easy way to
 * verify its behaviour. It generates the raw data for the plots in the readme
 * and spoiler documents.
 *
 * Thus, the following code contains spoilers!
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BINS 1024
#define MAXPEAK 14
#define HIST_RUNS 1000
#define AMOUNT 300000

struct material {
	int	npeak;
	int	x[MAXPEAK];
	int	n[MAXPEAK];
} refmaterial, wrongmaterial;

double testdata[BINS], refdata[BINS], moredata[3][BINS];

int match(double *test, double *reference, int bins, double threshold);

void generate_material(struct material *m) {
	int i;

	m->npeak = MAXPEAK - (rand() % 5);
	for(i = 0; i < m->npeak; i++) {
		m->x[i] = rand() % BINS;
		m->n[i] = 5 + (rand() % ((i & 1)? 20 : 40));
	}
}

void initnoise(double *spectrum) {
	int i;

	for(i = 0; i < BINS; i++) {
		spectrum[i] = rand() % 20;
	}
}

void generate(double *spectrum, struct material *m, int amount) {
	int h, j, k = 0, x;

	while(amount > 0) {
		k = (k + 1) % m->npeak;
		for(j = 0; j < m->n[k] && amount > 0; j++) {
			for(h = 0; h < 5; h++) {
				x = -BINS + rand() % BINS + rand() % BINS;
				if(x < 0) x = -x;
				if(x < m->x[k]) {
					spectrum[x]++;
					amount--;
				}
			}
			x = m->x[k];
			for(h = 0; h < 20; h++) {
				if(k & 1) {
					x += (rand() % 21) - 10;
				} else {
					x += (rand() % 3) - 1;
				}
			}
			if(x >= 0 && x < BINS) {
				spectrum[x]++;
				amount--;
			}
		}
	}
}

int find_t(double *testdata, double *refdata) {
	int t, tmin, tmax;

	tmin = -400;
	tmax = 396;
	while(tmin < tmax) {
		t = (tmin + tmax) / 2;
		if(match(testdata, refdata, BINS, (double) t / 400)) {
			tmin = t + 1;
		} else {
			tmax = t - 1;
		}
	}
	
	return round((double) tmin / 4);
}

static void smoothen(double *dest, double *src) {
	double sum;
	int i, j;
	for(i = 0; i < BINS; i++) {
		sum = 0;
		for(j = 0; j < 16 && i + j < BINS; j++) sum += src[i + j];
		dest[i] = sum / 16;
	}
}

static void differentiate(double *dest, double *src) {
	int i;

	for(i = 0; i < BINS - 1; i++) {
		dest[i] = src[i + 1] - src[i];
	}
	dest[BINS - 1] = 0;
}

void histogram(int plot) {
	int k, t;
	int hist[3][200];

	memset(hist, 0, sizeof(hist));
	for(k = (plot >= 5)? 2 * HIST_RUNS : 0; k < ((plot >= 5)? 3 : 2) * HIST_RUNS; k++) {
		generate_material(&refmaterial);
		generate_material(&wrongmaterial);
		initnoise(refdata);
		initnoise(testdata);
		generate(refdata, &refmaterial, AMOUNT);
		if(k < HIST_RUNS) {
			generate(testdata, &refmaterial, AMOUNT);
		} else if(k < 2 * HIST_RUNS) {
			generate(testdata, &wrongmaterial, AMOUNT);
		} else {
			generate(testdata, &refmaterial, AMOUNT / 10);
			wrongmaterial.npeak = 1;
			wrongmaterial.x[0] = BINS * 3 / 4;
			wrongmaterial.n[0] = 100;
			generate(testdata, &wrongmaterial, AMOUNT - AMOUNT / 10);
		}
		t = find_t(testdata, refdata);
		hist[k / HIST_RUNS][t + 100]++;
	}

	for(t = ((plot == 4)? 50 : 100); t < 200; t++) {
		if(plot == 4) {
			printf("%d %d\n", hist[0][t], hist[1][t]);
		} else if(plot == 5) {
			printf("%d\n", hist[2][t]);
		}
	}
}

int main(int argc, char **argv) {
	int i, t, plot;
	volatile double d;

	if(argc != 2) {
		fprintf(stderr, "Usage: %s <plot-id>\n", argv[0]);
		exit(1);
	}

	srand(45);

	plot = strtol(argv[1], 0, 10);

	switch(plot) {
	case 0:
	case 1:
	case 2:
	case 3:
	case 11:
		generate_material(&refmaterial);
		initnoise(refdata);
		generate(refdata, &refmaterial, AMOUNT);
		smoothen(moredata[0], refdata);
		differentiate(moredata[1], moredata[0]);
		smoothen(moredata[2], moredata[1]);
		if(plot == 11) {
			for(i = 0; i < BINS; i++) {
				printf("%f\n", ((float *) moredata[2])[i]);
			}
		} else {
			for(i = 0; i < BINS; i++) {
				if(plot == 0) {
					printf("%lf\n", refdata[i]);
				} else {
					printf("%lf\n", moredata[plot - 1][i]);
				}
			}
		}
		break;
	case 4:
	case 5:
		histogram(plot);
		break;
	case 6:
	case 7:
	case 8:
	case 9:
	case 10:
		generate_material(&refmaterial);
		initnoise(refdata);
		generate(refdata, &refmaterial, AMOUNT);
		do {
			generate_material(&wrongmaterial);
			initnoise(testdata);
			generate(testdata, &refmaterial, AMOUNT / 10);
			wrongmaterial.npeak = 1;
			wrongmaterial.x[0] = BINS * 3 / 4;
			wrongmaterial.n[0] = 100;
			generate(testdata, &wrongmaterial, AMOUNT - AMOUNT / 10);
			t = find_t(testdata, refdata);
			fprintf(stderr, "t = %d\n", t);
		} while(t < 65);
		smoothen(moredata[0], testdata);
		differentiate(moredata[1], moredata[0]);
		smoothen(moredata[2], moredata[1]);
		if(plot == 10) {
			for(i = 0; i < BINS; i++) {
				printf("%f\n", ((float *) moredata[2])[i]);
			}
		} else {
			for(i = 0; i < BINS; i++) {
				if(plot == 6) {
					printf("%lf\n", testdata[i]);
				} else {
					printf("%lf\n", moredata[plot - 7][i]);
				}
			}
		}
		break;
	case 12:
		for(i = -1000; i <= 1000; i++) {
			d = i;
			printf("%lf %f\n", d, ((float *) &d)[1]);
		}
		break;
	default:
		fprintf(stderr, "Unknown plot id\n");
		exit(1);
	}

	return 0;
}
